// Raw data from a sensor.
struct raw_measure
{
	int m_meter_id;

	uint8_t m_high_bits;
	uint8_t m_low_bits;
};

// Type of input for validation stage with raw data from a sensor.
struct raw_value
{
	raw_measure m_data;
};

// Type of input for conversion stage with valid raw data from a sensor.
struct valid_raw_value
{
	raw_measure m_data;
};

// Data from a sensor after conversion to Celsius degrees.
struct calculated_measure
{
	int m_meter_id;

	float m_measure;
};

// The type for result of conversion stage with converted data from a sensor.
struct sensor_value
{
	calculated_measure m_data;
};

// Type with value which could mean a dangerous level of temperature.
struct suspicious_value
{
	calculated_measure m_data;
};

// Type with information about detected dangerous situation.
struct alarm_detected
{
	int m_meter_id;
};

//
// The first stage of a pipeline. Validation of raw data from a sensor.
//
// Returns valid_raw_value or nothing if value is invalid.
//
stage_result_t< valid_raw_value >
validation( const raw_value & v )
{
	if( 0x7 >= v.m_data.m_high_bits )
		return make_result< valid_raw_value >( v.m_data );
	else
		return make_empty< valid_raw_value >();
}

//
// The second stage of a pipeline. Conversion from raw data to a value
// in Celsius degrees.
//
stage_result_t< sensor_value >
conversion( const valid_raw_value & v )
{
	return make_result< sensor_value >(
		calculated_measure{ v.m_data.m_meter_id,
			0.5f * ((static_cast< uint16_t >( v.m_data.m_high_bits ) << 8) +
				v.m_data.m_low_bits) } );
}

//
// Simulation of the data archiving.
//
void
archiving( const sensor_value & v )
{
	clog << "archiving (" << v.m_data.m_meter_id << ","
		<< v.m_data.m_measure << ")" << endl;
}

//
// Simulation of the data distribution.
//
void
distribution( const sensor_value & v )
{
	clog << "distributing (" << v.m_data.m_meter_id << ","
		<< v.m_data.m_measure << ")" << endl;
}

//
// The first stage of a child pipeline at third level of the main pipeline.
//
// Checking for to high value of the temperature.
//
// Returns suspicious_value message or nothing.
//
stage_result_t< suspicious_value >
range_checking( const sensor_value & v )
{
	if( v.m_data.m_measure >= 45.0f )
		return make_result< suspicious_value >( v.m_data );
	else
		return make_empty< suspicious_value >();
}

//
// The next stage of a child pipeline.
//
// Checks for two suspicious_value-es in 25ms time window.
//
class alarm_detector
{
	using clock = chrono::steady_clock;

public :
	stage_result_t< alarm_detected >
	operator()( const suspicious_value & v )
	{
		if( m_previous )
			if( *m_previous + chrono::milliseconds(25) > clock::now() )
			{
				m_previous = nullopt;
				return make_result< alarm_detected >( v.m_data.m_meter_id );
			}

		m_previous = clock::now();
		return make_empty< alarm_detected >();
	}

private :
	optional< clock::time_point > m_previous;
};

//
// One of last stages of a child pipeline.
// Imitates beginning of the alarm processing.
//
void
alarm_initiator( const alarm_detected & v )
{
	clog << "=== alarm (" << v.m_meter_id << ") ===" << endl;
}

//
// Another of last stages of a child pipeline.
// Imitates distribution of the alarm.
//
void
alarm_distribution( ostream & to, const alarm_detected & v )
{
	to << "alarm_distribution (" << v.m_meter_id << ")" << endl;
}

