//
// All messages will be passed as SObjectizer-messages.
// It means that they will be created as dynamically allocated objects.
// The name stage_result_t will be used as alias for smart-pointer to
// dynamically created message.
//
// The name stage_result_t means that messages to be passed will be
// returned as stage processing result.
//
template< typename M >
using stage_result_t = message_holder_t< M >;

//
// Just a helper function for creating new message instance.
// Something like std::make_shared or std::make_unique.
//
template< typename M, typename... Args >
stage_result_t< M >
make_result( Args &&... args )
{
	return stage_result_t< M >::make(forward< Args >(args)...);
}

//
// Just a helper function for the case when there is no processing result
// (stage has to return nothing).
//
template< typename M >
stage_result_t< M >
make_empty()
{
	return stage_result_t< M >();
}

//
// We have to deal with two types of stage handlers:
// - intermediate handlers which will return some result (e.g. some new
//   message);
// - terminal handlers which can return nothing (e.g. void instead of
//   stage_result_t<M>);
//
// This template with specialization defines `input` and `output`
// aliases for both cases.
//
template< typename In, typename Out >
struct handler_traits_t
{
	using input = In;
	using output = stage_result_t< Out >;
};

template< typename In >
struct handler_traits_t< In, void >
{
	using input = In;
	using output = void;
};

//
// A template for repesentation of one stage handler.
//
// The handler could be specified as free function pointer, as stateful
// object with operator() or as lambda function. The actual handler will
// be stored as instance of std::function.
//
template< typename In, typename Out >
class stage_handler_t
{
public :
	using traits = handler_traits_t< In, Out >;
	using func_type = function< typename traits::output(const typename traits::input &) >;

	stage_handler_t( func_type handler )
		: m_handler( move(handler) )
	{}

	template< typename Callable >
	stage_handler_t( Callable handler ) : m_handler( handler ) {}

	typename traits::output
	operator()( const typename traits::input & a ) const
	{
		return m_handler( a );
	}

private :
	func_type m_handler;
};

//
// An agent which will be used as intermediate or terminal pipeline stage.
// It will receive input message, call the stage handler and pass
// handler result to the next stage (if any).
//
template< typename In, typename Out >
class a_stage_point_t final : public agent_t
{
public :
	a_stage_point_t(
		context_t ctx,
		stage_handler_t< In, Out > handler,
		mbox_t next_stage )
		:	agent_t{ ctx }
		,	m_handler{ move( handler ) }
		,	m_next{ move(next_stage) }
	{}

	void so_define_agent() override
	{
		if( m_next )
			// Because there is the next stage the appropriate
			// message handler will be used.
			so_subscribe_self().event( [=]( const In & evt ) {
					auto r = m_handler( evt );
					if( r )
						so_5::send( m_next, r );
				} );
		else
			// There is no next stage. A very simple message handler
			// will be used for that case.
			so_subscribe_self().event( [=]( const In & evt ) {
					m_handler( evt );
				} );
	}

private :
	const stage_handler_t< In, Out > m_handler;
	const mbox_t m_next;
};

//
// A specialization of a_stage_point_t for the case of terminal stage of
// a pipeline. This type will be used for stage handlers with void
// return type.
//
template< typename In >
class a_stage_point_t< In, void > final : public agent_t
{
public :
	a_stage_point_t(
		context_t ctx,
		stage_handler_t< In, void > handler,
		mbox_t next_stage )
		:	agent_t{ ctx }
		,	m_handler{ move( handler ) }
	{
		if( next_stage )
			throw std::runtime_error( "sink point cannot have next stage" );
	}

	void so_define_agent() override
	{
		so_subscribe_self().event( [=]( const In & evt ) {
				m_handler( evt );
			} );
	}

private :
	const stage_handler_t< In, void > m_handler;
};

//
// A special mbox for broadcasting of a message to a set of destination
// mboxes.
//
using broadcast_mbox_t = so_5::extra::mboxes::broadcast::fixed_mbox_template_t<>;

//
// An alias for functional object to build necessary agent for a
// pipeline stage.
//
using stage_builder_t = function< mbox_t(coop_t &, mbox_t) >;

//
// Description of one pipeline stage.
//
template< typename In, typename Out >
struct stage_t
{
	stage_builder_t m_builder;
};

//
// A bunch of definitions related to detection of handler argument
// type and return value.
//

// 
// Helper type for `arg_type` and `result_type` alises definition.
//
template< typename R, typename A >
struct callable_traits_typedefs_t
{
	using arg_type = A;
	using result_type = R;
};

//
// Helper type for dealing with stateful objects with operator()
// (they could be user-defined objects or generated by compiler
// like lambdas).
//
template< typename T >
struct lambda_traits_t;

template< typename M, typename A, typename T >
struct lambda_traits_t< stage_result_t< M >(T::*)(const A &) const >
	:	public callable_traits_typedefs_t< M, A >
{};

template< typename A, typename T >
struct lambda_traits_t< void (T::*)(const A &) const >
	:	public callable_traits_typedefs_t< void, A >
{};

template< typename M, typename A, typename T >
struct lambda_traits_t< stage_result_t< M >(T::*)(const A &) >
	:	public callable_traits_typedefs_t< M, A >
{};

template< typename A, typename T >
struct lambda_traits_t< void (T::*)(const A &) >
	:	public callable_traits_typedefs_t< void, A >
{};

//
// Main type for definition of `arg_type` and `result_type` aliases.
// With specialization for various cases.
//
template< typename T >
struct callable_traits_t
	:	public lambda_traits_t< decltype(&T::operator()) >
{};

template< typename M, typename A >
struct callable_traits_t< stage_result_t< M >(*)(const A &) >
	:	public callable_traits_typedefs_t< M, A >
{};

template< typename A >
struct callable_traits_t< void(*)(const A &) >
	:	public callable_traits_typedefs_t< void, A >
{};

//
// Main function for defining intermediate of terminal stage of a pipeline.
//
template<
	typename Callable,
	typename In = typename callable_traits_t< Callable >::arg_type,
	typename Out = typename callable_traits_t< Callable >::result_type >
stage_t< In, Out >
stage( Callable handler )
{
	stage_builder_t builder{
			[h = std::move(handler)](
				coop_t & coop,
				mbox_t next_stage) -> mbox_t
			{
				return coop.make_agent< a_stage_point_t<In, Out> >(
						std::move(h),
						std::move(next_stage) )
					->so_direct_mbox();
			}
	};

	return { std::move(builder) };
}

//
// Serie of helper functions for building description for
// `broadcast` stage.
//
// Those functions are used for collecting
// `builders` functions for every child pipeline.
//
// Please note that this functions checks that each child pipeline has the
// same In type.
//
template< typename In, typename Out, typename... Rest >
void
move_sink_builder_to(
	vector< stage_builder_t > & receiver,
	stage_t< In, Out > && first,
	Rest &&... rest )
{
	receiver.emplace_back( move( first.m_builder ) );
	if constexpr( 0u != sizeof...(rest) )
		move_sink_builder_to<In>( receiver, forward< Rest >(rest)... );
}

template< typename In, typename Out, typename... Rest >
vector< stage_builder_t >
collect_sink_builders( stage_t< In, Out > && first, Rest &&... stages )
{
	vector< stage_builder_t > receiver;
	receiver.reserve( 1 + sizeof...(stages) );
	move_sink_builder_to<In>(
			receiver,
			move(first),
			std::forward<Rest>(stages)... );

	return receiver;
}

//
// Main helper function for building `broadcast` stage.
//
template< typename In, typename Out, typename... Rest >
stage_t< In, void >
broadcast( stage_t< In, Out > && first, Rest &&... stages )
{
	stage_builder_t builder{
		[broadcasts = collect_sink_builders(
				move(first), forward< Rest >(stages)...)]
		( coop_t & coop, mbox_t ) -> mbox_t
		{
			vector< mbox_t > mboxes;
			mboxes.reserve( broadcasts.size() );

			for( const auto & b : broadcasts )
				mboxes.emplace_back( b( coop, mbox_t{} ) );

			return broadcast_mbox_t::make( coop.environment(), std::move(mboxes) );
		}
	};

	return { std::move(builder) };
}

//
// Helper `operator|` for continuation of a pipeline definition.
//
template< typename In, typename Out1, typename Out2 >
stage_t< In, Out2 >
operator|(
	stage_t< In, Out1 > && prev,
	stage_t< Out1, Out2 > && next )
{
	return {
		stage_builder_t{
			[prev, next]( coop_t & coop, mbox_t next_stage ) -> mbox_t
			{
				auto m = next.m_builder( coop, std::move(next_stage) );
				return prev.m_builder( coop, std::move(m) );
			}
		}
	};
}

//
// Main function for creation of all pipeline-related stuff.
//
template< typename In, typename Out, typename... Args >
mbox_t
make_pipeline(
	// SObjectizer Environment to work in.
	so_5::environment_t & env,
	// Definition of a pipeline.
	stage_t< In, Out > && sink,
	// Optional args to be passed to make_coop() function.
	Args &&... args )
{
	auto coop = env.make_coop( forward< Args >(args)... );

	auto mbox = sink.m_builder( *coop, mbox_t{} );

	env.register_coop( move(coop) );

	return mbox;
}

