#include <iostream>
#include <cstdint>
#include <optional>

#include <so_5/all.hpp>
#include <so_5_extra/mboxes/broadcast.hpp>

using namespace std;

using namespace so_5;

/*
 * The first part.
 *
 * Definition of low-level pipeline implementation details.
 */

#include "../common/pipeline_parts.ipp"

//
// Main function for defining intermediate of terminal stage of a pipeline
// with the specified dispatcher binder.
//
template<
	typename Callable,
	typename In = typename callable_traits_t< Callable >::arg_type,
	typename Out = typename callable_traits_t< Callable >::result_type >
stage_t< In, Out >
stage( 
	disp_binder_shptr_t disp_binder,
	Callable handler )
{
	stage_builder_t builder{
			[binder = std::move(disp_binder), h = std::move(handler)](
				coop_t & coop,
				mbox_t next_stage) -> mbox_t
			{
				return coop.make_agent_with_binder< a_stage_point_t<In, Out> >(
						std::move(binder),
						std::move(h),
						std::move(next_stage) )
					->so_direct_mbox();
			}
	};

	return { std::move(builder) };
}


/*
 * The second part.
 *
 * Definition of messages to be processed by a pipeline and
 * the message processing code.
 */

#include "../common/application_logic.ipp"

int main()
{
	// Launch SObjectizer in a separate thread.
	wrapped_env_t sobj;

	/*
	 * The third part.
	 *
	 * Definition of message processing pipeline and imitation of
	 * several measures from a sensor.
	 */

	// An active_obj dispatcher to be used for some stages.
	auto ao_disp = disp::active_obj::make_dispatcher( sobj.environment() );

	// Make a pipeline.
	auto pipeline = make_pipeline( sobj.environment(),
			stage(validation) | stage(conversion) | broadcast(
				stage(ao_disp.binder(), archiving),
				stage(ao_disp.binder(), distribution),
				stage(range_checking) | stage(alarm_detector{}) | broadcast(
					stage(ao_disp.binder(), alarm_initiator),
					stage(ao_disp.binder(), []( const alarm_detected & v ) {
							alarm_distribution( cerr, v );
						} )
					)
				),
			disp::one_thread::make_dispatcher( sobj.environment() ).binder() );

	// Send messages to a pipeline in a loop with 10ms delays.
	for( uint8_t i = 0; i < static_cast< uint8_t >(250); i += 10 )
	{
		send< raw_value >(
				pipeline,
				raw_measure{ 0, 0, i } );
		std::this_thread::sleep_for( chrono::milliseconds{10} );
	}
}

